# Copyright 2014-2020 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

# Standard modules
import logging
import textwrap
import os

from qtpy import QtCore, QtWidgets, QtGui, PYQT_VERSION, PYSIDE_VERSION

# Custom modules
from . import components
from .helpers.profiling import StopWatch

# Logging
# Setup the logger to use a different name than the file name
log = logging.getLogger("grc.application")


class Application(QtWidgets.QApplication):
    """
    This is the main QT application for GRC.
    It handles setting up the application components and actions and handles communication between different components in the system.
    """

    def __init__(self, settings, platform, file_path):
        # Note. Logger must have the correct naming convention to share handlers
        log.debug("__init__")
        self.settings = settings
        self.platform = platform
        config = platform.config

        self.qsettings = QtCore.QSettings(config.gui_prefs_file, QtCore.QSettings.IniFormat)
        log.debug(f"Using QSettings from {config.gui_prefs_file}")
        os.environ["QT_SCALE_FACTOR"] = self.qsettings.value('appearance/qt_scale_factor', "1.0", type=str)

        log.debug("Creating QApplication instance")
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts, True)
        QtWidgets.QApplication.__init__(self, settings.argv)

        self.theme = "light"
        if self.qsettings.value("appearance/theme", "dark") == "dark":
            try:
                import qdarkstyle

                self.setStyleSheet(qdarkstyle.load_stylesheet())
                self.theme = "dark"
            except ImportError:
                log.warning("Did not find QDarkstyle. Dark mode disabled")

        # Load the main view class and initialize QMainWindow
        log.debug("ARGV - {0}".format(settings.argv))
        log.debug("INSTALL_DIR - {0}".format(settings.path.INSTALL))

        # Global signals
        self.signals = {}

        # Setup the main application window
        log.debug("Creating main application window")
        stopwatch = StopWatch()
        self.MainWindow = components.MainWindow(file_path)
        stopwatch.lap("mainwindow")
        level_str = self.qsettings.value('grc/console_log_level', "info", type=str)
        if level_str == "info":
            self.Console = components.Console(logging.INFO)
        else:  # level_str == "debug"
            self.Console = components.Console(logging.DEBUG)
        stopwatch.lap("console")
        self.BlockLibrary = components.BlockLibrary()
        stopwatch.lap("blocklibrary")
        # self.DocumentationTab = components.DocumentationTab()
        # stopwatch.lap('documentationtab')
        self.WikiTab = components.WikiTab()
        stopwatch.lap("wikitab")
        self.VariableEditor = components.VariableEditor()
        stopwatch.lap("variable_editor")
        self.VariableEditor.set_scene(self.MainWindow.currentFlowgraphScene)
        self.VariableEditor.all_editor_actions.connect(self.MainWindow.handle_editor_action)
        self.MainWindow.ExampleBrowser.set_library(self.BlockLibrary)

        # Debug times
        log.debug(
            "Loaded MainWindow controller - {:.4f}s".format(
                stopwatch.elapsed("mainwindow")
            )
        )
        log.debug(
            "Loaded Console component - {:.4f}s".format(stopwatch.elapsed("console"))
        )
        log.debug(
            "Loaded BlockLibrary component - {:.4}s".format(
                stopwatch.elapsed("blocklibrary")
            )
        )
        # log.debug("Loaded DocumentationTab component - {:.4}s".format(stopwatch.elapsed("documentationtab")))

        # Print Startup information once everything has loaded
        self.Console.enable()

        paths = "\n\t".join(config.block_paths)
        welcome = (
            f"<<< Welcome to {config.name} {config.version} >>>\n\n"
            f"{('PyQt ' + PYQT_VERSION) if PYQT_VERSION else ('PySide ' + PYSIDE_VERSION)}\n"
            f"GUI preferences file: {self.qsettings.fileName()}\n"
            f"Block paths:\n\t{paths}\n"
            f"Using {QtGui.QIcon.themeName()} icon theme\n"
        )
        log.info(textwrap.dedent(welcome))

        log.debug(f'devicePixelRatio {self.MainWindow.screen().devicePixelRatio()}')

        if (self.qsettings.value("appearance/theme", "dark") == "dark") and (self.theme == "light"):
            log.warning("Could not apply dark theme. Is QDarkStyle installed?")

    # Global registration functions
    #  - Handles the majority of child controller interaciton

    def registerSignal(self, signal):
        pass

    def registerDockWidget(self, widget, location=0):
        """Allows child controllers to register a widget that can be docked in the main window"""
        # TODO: Setup the system to automatically add new "Show <View Name>" menu items when a new
        # dock widget is added.
        log.debug(
            "Registering widget ({0}, {1})".format(widget.__class__.__name__, location)
        )
        self.MainWindow.registerDockWidget(location, widget)

    def registerMenu(self, menu):
        """Allows child controllers to register an a menu rather than just a single action"""
        # TODO: Need to have several capabilities:
        #  - Menu's need the ability to have a priority for ordering
        #  - When registering, the controller needs to specific target menu
        #  - Automatically add seporators (unless it is the first or last items)
        #  - Have the ability to add it as a sub menu
        #  - MainWindow does not need to call register in the app controller. It can call directly
        #  - Possibly view sidebars and toolbars as submenu
        #  - Have the ability to create an entirely new menu
        log.debug("Registering menu ({0})".format(menu.title()))
        self.MainWindow.registerMenu(menu)

    def registerAction(self, action, menu):
        """Allows child controllers to register a global action shown in the main window"""
        pass

    def run(self):
        """Launches the main QT event loop"""
        # Show the main window after everything is initialized.
        self.MainWindow.show()
        return self.exec_()



# Copyright 2014-2020 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

# Standard modules
import logging
import functools

from qtpy import QtGui, QtCore, QtWidgets, QT6
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication

from itertools import count

# Custom modules
from ....core.base import Element
from .connection import DummyConnection
from .port import GUIPort
from ... import base
from ....core.FlowGraph import FlowGraph as CoreFlowgraph
from ... import Utils
from ...external_editor import ExternalEditor

from .block import GUIBlock
from .connection import GUIConnection

# Logging
log = logging.getLogger(f"grc.application.{__name__}")

DEFAULT_MAX_X = 400
DEFAULT_MAX_Y = 300


class Flowgraph(CoreFlowgraph):
    def __init__(self, gui, platform, *args, **kwargs):
        self.gui = gui
        CoreFlowgraph.__init__(self, platform)


class FlowgraphScene(QtWidgets.QGraphicsScene, base.Component):
    itemMoved = QtCore.Signal([QtCore.QPointF])
    newElement = QtCore.Signal([Element])
    deleteElement = QtCore.Signal([Element])
    blockPropsChange = QtCore.Signal([Element])

    def __init__(self, view, platform, *args, **kwargs):
        self.core = Flowgraph(self, platform)
        super(FlowgraphScene, self).__init__()
        self.setParent(view)
        self.view = view
        self.isPanning = False
        self.mousePressed = False
        self.moving_blocks = False

        self.dummy_arrow = None
        self.start_port = None
        self.end_port = None
        self._elements_to_draw = []
        self._external_updaters = {}

        self.qsettings = QApplication.instance().qsettings

        if QT6:
            self.undoStack = QtGui.QUndoStack(self)
        else:
            self.undoStack = QtWidgets.QUndoStack(self)
        self.undoAction = self.undoStack.createUndoAction(self, "Undo")
        self.redoAction = self.undoStack.createRedoAction(self, "Redo")

        self.filename = None

        self.clickPos = None

        self.saved = False
        self.save_allowed = True

    def set_saved(self, saved):
        self.saved = saved

    def update(self):
        """
        Call the top level rewrite and validate.
        Call the top level create labels and shapes.
        """
        self.core.rewrite()
        self.core.validate()
        for block in self.core.blocks:
            block.gui.create_shapes_and_labels()
        self.update_elements_to_draw()
        self.app.VariableEditor.update_gui(self.core.blocks)

    def update_elements_to_draw(self):
        # hide_disabled_blocks = Actions.TOGGLE_HIDE_DISABLED_BLOCKS.get_active()
        hide_disabled_blocks = False
        # hide_variables = Actions.TOGGLE_HIDE_VARIABLES.get_active()
        hide_variables = False

        def draw_order(elem):
            return elem.gui.isSelected(), elem.is_block, elem.enabled

        elements = sorted(self.core.get_elements(), key=draw_order)
        del self._elements_to_draw[:]

        for element in elements:
            if hide_disabled_blocks and not element.enabled:
                continue  # skip hidden disabled blocks and connections
            if hide_variables and (element.is_variable or element.is_import):
                continue  # skip hidden disabled blocks and connections
            self._elements_to_draw.append(element)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def decode_data(self, bytearray):
        data = []
        item = {}
        ds = QtCore.QDataStream(bytearray)
        while not ds.atEnd():
            row = ds.readInt32()
            column = ds.readInt32()
            map_items = ds.readInt32()
            for i in range(map_items):
                key = ds.readInt32()
                value = QtCore.QVariant()
                ds >> value
                item[Qt.ItemDataRole(key)] = value
            data.append(item)
        return data

    def _get_unique_id(self, base_id=""):
        """
        Get a unique id starting with the base id.

        Args:
            base_id: the id starts with this and appends a count

        Returns:
            a unique id
        """
        block_ids = set(b.name for b in self.core.blocks)
        for index in count():
            block_id = "{}_{}".format(base_id, index)
            if block_id not in block_ids:
                break
        return block_id

    def dropEvent(self, event):
        QtWidgets.QGraphicsScene.dropEvent(self, event)
        if event.mimeData().hasUrls:
            data = event.mimeData()
            if data.hasFormat("application/x-qabstractitemmodeldatalist"):
                bytearray = data.data("application/x-qabstractitemmodeldatalist")
                data_items = self.decode_data(bytearray)

                # Find block in tree so that we can pull out label
                block_key = data_items[0][QtCore.Qt.UserRole].value()

                # Add block of this key at the cursor position
                cursor_pos = event.scenePos()
                pos = (cursor_pos.x(), cursor_pos.y())

                self.add_block(block_key, pos)

                event.setDropAction(Qt.CopyAction)
                event.accept()
            else:
                return QtGui.QStandardItemModel.dropMimeData(
                    self, data, action, row, column, parent
                )
        else:
            event.ignore()

    def add_block(self, block_key, pos=(0, 0)):
        block = self.platform.blocks[block_key]
        # Pull out its params (keep in mind we still havent added the dialog box that lets you change param values so this is more for show)
        params = []
        for (
            p
        ) in (
            block.parameters_data
        ):  # block.parameters_data is a list of dicts, one per param
            if "label" in p:  # for now let's just show it as long as it has a label
                key = p["label"]
                value = p.get("default", "")  # just show default value for now
                params.append((key, value))

        id = self._get_unique_id(block_key)

        c_block = self.core.new_block(block_key)
        g_block = c_block.gui
        c_block.states["coordinate"] = pos
        g_block.setPos(*pos)
        c_block.params["id"].set_value(id)
        self.addItem(g_block)
        g_block.moveToTop()
        self.update()
        self.newElement.emit(c_block)
        return id

    def selected_blocks(self) -> list[GUIBlock]:
        blocks = []
        for item in self.selectedItems():
            if isinstance(item, GUIBlock):
                blocks.append(item)
        return blocks

    def selected_connections(self):
        conns = []
        for item in self.selectedItems():
            if isinstance(item, GUIConnection):
                conns.append(item)
        return conns

    def delete_selected(self):
        for item in self.selectedItems():
            self.remove_element(item)

    def select_all(self):
        for block in self.core.blocks:
            block.gui.setSelected(True)
        for conn in self.core.connections:
            conn.gui.setSelected(True)

    def rotate_selected(self, rotation):
        """
        Rotate the selected blocks by multiples of 90 degrees.
        Args:
            rotation: the rotation in degrees
        Returns:
            true if changed, otherwise false.
        """
        selected_blocks = self.selected_blocks()
        if not any(selected_blocks):
            return False
        # initialize min and max coordinates
        min_x, min_y = max_x, max_y = selected_blocks[0].x(), selected_blocks[0].y()
        # rotate each selected block, and find min/max coordinate
        for selected_block in selected_blocks:
            selected_block.rotate(rotation)
            # update the min/max coordinate
            x, y = selected_block.x(), selected_block.y()
            min_x, min_y = min(min_x, x), min(min_y, y)
            max_x, max_y = max(max_x, x), max(max_y, y)
        # calculate center point of selected blocks
        ctr_x, ctr_y = (max_x + min_x) / 2, (max_y + min_y) / 2
        # rotate the blocks around the center point
        for selected_block in selected_blocks:
            x, y = selected_block.x(), selected_block.y()
            x, y = Utils.get_rotated_coordinate((x - ctr_x, y - ctr_y), rotation)
            selected_block.setPos(x + ctr_x, y + ctr_y)
        return True

    def mousePressEvent(self, event):
        g_item = self.itemAt(event.scenePos(), QtGui.QTransform())
        if not g_item:  # Nothing selected
            event.ignore()
            return

        self.clickPos = event.scenePos()
        self.moving_blocks = False

        if g_item and not isinstance(g_item, DummyConnection):
            c_item = g_item.core
            if c_item.is_block:
                self.moving_blocks = True
            elif c_item.is_port:
                if c_item.is_source:
                    self.start_port = g_item
                elif c_item.is_sink:
                    self.end_port = g_item
        if event.button() == Qt.LeftButton:
            self.mousePressed = True
            super(FlowgraphScene, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self.mousePressed:
            if not self.dummy_arrow and not self.moving_blocks and self.start_port:
                self.dummy_arrow = DummyConnection(self, self.start_port.connection_point, self.clickPos)
                self.addItem(self.dummy_arrow)
        self.view.setSceneRect(self.itemsBoundingRect())
        if self.dummy_arrow:
            self.dummy_arrow.update(event.scenePos())

        if self.mousePressed and self.isPanning:
            new_pos = event.pos()
            self.dragPos = new_pos
            event.accept()
        else:
            super(FlowgraphScene, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self.dummy_arrow:  # We are currently dragging a DummyConnection
            self.removeItem(self.dummy_arrow)
            g_item = self.itemAt(event.scenePos(), QtGui.QTransform())
            if isinstance(g_item, GUIPort):
                c_item = g_item.core
                if g_item != self.start_port:
                    log.debug("Created connection (drag)")
                    new_con = self.core.connect(self.start_port.core, c_item)
                    self.addItem(new_con.gui)
                    self.newElement.emit(new_con)
                    self.update()
                    self.start_port = None
                    self.end_port = None
            self.dummy_arrow = None
        else:
            if self.clickPos != event.scenePos() and self.moving_blocks:
                self.itemMoved.emit(event.scenePos() - self.clickPos)
            elif (self.start_port != None) and (self.end_port != None):
                log.debug("Created connection (click)")
                new_con = self.core.connect(self.start_port.core, self.end_port.core)
                self.addItem(new_con.gui)
                self.newElement.emit(new_con)
                self.update()
                self.start_port = None
                self.end_port = None
        self.mousePressed = False
        self.moving_blocks = False
        super(FlowgraphScene, self).mouseReleaseEvent(event)

    def createActions(self, actions):
        log.debug("Creating actions")

        """
        # File Actions
        actions['save'] = Action(Icons("document-save"), _("save"), self,
                                shortcut=Keys.New, statusTip=_("save-tooltip"))

        actions['clear'] = Action(Icons("document-close"), _("clear"), self,
                                         shortcut=Keys.Open, statusTip=_("clear-tooltip"))
        """

    def createMenus(self, actions, menus):
        log.debug("Creating menus")

    def createToolbars(self, actions, toolbars):
        log.debug("Creating toolbars")

    def import_data(self, data):
        self.core.import_data(data)
        for conn in self.core.connections:
            self.addItem(conn.gui)
        for block in self.core.blocks:
            self.addItem(block.gui)

    def getMaxZValue(self):
        z_values = []
        for block in self.core.blocks:
            z_values.append(block.gui.zValue())
        return max(z_values)

    def remove_element(self, element: GUIBlock):
        self.removeItem(element)
        self.core.remove_element(element.core)

    def get_extents(self):
        # show_comments = Actions.TOGGLE_SHOW_BLOCK_COMMENTS.get_active()
        show_comments = True

        def sub_extents():
            for element in self._elements_to_draw:
                yield element.get_extents()
                if element.is_block and show_comments and element.enabled:
                    yield element.get_extents_comment()

        extent = 10000000, 10000000, 0, 0
        cmps = (min, min, max, max)
        for sub_extent in sub_extents():
            extent = [cmp(xy, e_xy) for cmp, xy, e_xy in zip(cmps, extent, sub_extent)]
        return tuple(extent)

    def copy_to_clipboard(self):
        """
        Copy the selected blocks and connections into the clipboard.

        Returns:
            the clipboard
        """
        # get selected blocks
        g_blocks = list(self.selected_blocks())
        if not g_blocks:
            return None
        # calc x and y min
        x_min, y_min = g_blocks[0].core.states["coordinate"]
        for g_block in g_blocks:
            x, y = g_block.core.states["coordinate"]
            x_min = min(x, x_min)
            y_min = min(y, y_min)
        # get connections between selected blocks
        connections = list(
            filter(
                lambda c: c.source_block.gui in g_blocks and c.sink_block.gui in g_blocks,
                self.core.connections,
            )
        )
        clipboard = (
            (x_min, y_min),
            [g_block.core.export_data() for g_block in g_blocks],
            [c_connection.export_data() for c_connection in connections],
        )
        return clipboard

    def paste_from_clipboard(self, clipboard):
        """
        Paste the blocks and connections from the clipboard.

        Args:
            clipboard: the nested data of blocks, connections
        """
        self.clearSelection()
        (x_min, y_min), blocks_n, connections_n = clipboard
        """
        # recalc the position
        scroll_pane = self.drawing_area.get_parent().get_parent()
        h_adj = scroll_pane.get_hadjustment()
        v_adj = scroll_pane.get_vadjustment()
        x_off = h_adj.get_value() - x_min + h_adj.get_page_size() / 4
        y_off = v_adj.get_value() - y_min + v_adj.get_page_size() / 4

        if len(self.get_elements()) <= 1:
            x_off, y_off = 0, 0
        """
        x_off, y_off = 10, 10

        # create blocks
        pasted_blocks = {}
        for block_n in blocks_n:
            block_key = block_n.get("id")
            if block_key == "options":
                continue

            block_name = block_n.get("name")
            # Verify whether a block with this name exists before adding it
            if block_name in (blk.name for blk in self.core.blocks):
                block_n = block_n.copy()
                block_n["name"] = self._get_unique_id(block_name)

            block = self.core.new_block(block_key)
            if not block:
                continue  # unknown block was pasted (e.g. dummy block)

            block.import_data(**block_n)
            pasted_blocks[block_name] = block  # that is before any rename

            block.gui.moveBy(x_off, y_off)
            self.addItem(block.gui)
            block.gui.moveToTop()
            block.gui.setSelected(True)
            """
            while any(Utils.align_to_grid(block.states['coordinate']) == Utils.align_to_grid(other.states['coordinate'])
                      for other in self.blocks if other is not block):
                block.moveBy(Constants.CANVAS_GRID_SIZE,
                           Constants.CANVAS_GRID_SIZE)
                # shift all following blocks
                x_off += Constants.CANVAS_GRID_SIZE
                y_off += Constants.CANVAS_GRID_SIZE
            """

        # update before creating connections
        self.update()
        # create connections
        for src_block, src_port, dst_block, dst_port in connections_n:
            source = pasted_blocks[src_block].get_source(src_port)
            sink = pasted_blocks[dst_block].get_sink(dst_port)
            connection = self.core.connect(source, sink)
            self.addItem(connection.gui)
            connection.gui.setSelected(True)

    def itemsBoundingRect(self):
        rect = QtWidgets.QGraphicsScene.itemsBoundingRect(self)
        return QtCore.QRectF(0.0, 0.0, rect.right(), rect.bottom())

    def install_external_editor(self, param, parent=None):
        target = (param.parent_block.name, param.key)

        if target in self._external_updaters:
            editor = self._external_updaters[target]
        else:
            editor = self.qsettings.value("grc/editor", "")
            if not editor:
                return
            updater = functools.partial(
                self.handle_external_editor_change, target=target)
            editor = self._external_updaters[target] = ExternalEditor(
                editor=editor,
                name=target[0], value=param.get_value(),
                callback=updater
            )
            editor.start()
        try:
            editor.open_editor()
        except Exception as e:
            # Problem launching the editor. Need to select a new editor.
            log.error('Error opening an external editor. Please select a different editor.\n')
            # Reset the editor to force the user to select a new one.
            self.parent_platform.config.editor = ''
            self.remove_external_editor(target=target)

    def remove_external_editor(self, target=None, param=None):
        if target is None:
            target = (param.parent_block.name, param.key)
        if target in self._external_updaters:
            self._external_updaters[target].stop()
            del self._external_updaters[target]

    def handle_external_editor_change(self, new_value, target):
        try:
            block_id, param_key = target
            self.core.get_block(block_id).params[param_key].set_value(new_value)

        except (IndexError, ValueError):  # block no longer exists
            self.remove_external_editor(target=target)
            return



# Copyright 2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# SPDX-License-Identifier: GPL-2.0-or-later
#


import itertools

from . import Block, register_build_in
from ._build import build_params


@register_build_in
class VirtualSink(Block):
    count = itertools.count()

    key = 'virtual_sink'
    label = 'Virtual Sink'
    flags = Block.flags
    flags.set('cpp')

    parameters_data = build_params(
        params_raw=[
            dict(label='Stream ID', id='stream_id', dtype='stream_id')],
        have_inputs=False, have_outputs=False, flags=flags, block_id=key
    )
    inputs_data = [dict(domain='stream', dtype='', direction='sink', id="0")]

    def __init__(self, parent, **kwargs):
        super(VirtualSink, self).__init__(parent, **kwargs)
        self.params['id'].hide = 'all'

    @property
    def stream_id(self):
        return self.params['stream_id'].value


@register_build_in
class VirtualSource(Block):
    count = itertools.count()

    key = 'virtual_source'
    label = 'Virtual Source'
    flags = Block.flags
    flags.set('cpp')

    parameters_data = build_params(
        params_raw=[
            dict(label='Stream ID', id='stream_id', dtype='stream_id')],
        have_inputs=False, have_outputs=False, flags=flags, block_id=key
    )
    outputs_data = [dict(domain='stream', dtype='',
                         direction='source', id="0")]

    def __init__(self, parent, **kwargs):
        super(VirtualSource, self).__init__(parent, **kwargs)
        self.params['id'].hide = 'all'

    @property
    def stream_id(self):
        return self.params['stream_id'].value



# Copyright 2008, 2009, 2015, 2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# SPDX-License-Identifier: GPL-2.0-or-later
#

import weakref

from .utils.descriptors import lazy_property


class Element(object):

    def __init__(self, parent=None):
        self._parent = weakref.ref(parent) if parent else lambda: None
        self._error_messages = []

    ##################################################
    # Element Validation API
    ##################################################
    def validate(self):
        """
        Validate this element and call validate on all children.
        Call this base method before adding error messages in the subclass.
        """
        for child in self.children():
            child.validate()

    def is_valid(self):
        """
        Is this element valid?

        Returns:
            true when the element is enabled and has no error messages or is bypassed
        """
        if not self.enabled or self.get_bypassed():
            return True
        return not next(self.iter_error_messages(), False)

    def add_error_message(self, msg):
        """
        Add an error message to the list of errors.

        Args:
            msg: the error message string
        """
        self._error_messages.append(msg)

    def get_error_messages(self):
        """
        Get the list of error messages from this element and all of its children.
        Do not include the error messages from disabled or bypassed children.
        Cleverly indent the children error messages for printing purposes.

        Returns:
            a list of error message strings
        """
        return [msg if elem is self else "{}:\n\t{}".format(elem, msg.replace("\n", "\n\t"))
                for elem, msg in self.iter_error_messages()]

    def iter_error_messages(self):
        """
        Iterate over error messages. Yields tuples of (element, message)
        """
        for msg in self._error_messages:
            yield self, msg
        for child in self.children():
            if not child.enabled or child.get_bypassed():
                continue
            for element_msg in child.iter_error_messages():
                yield element_msg

    def rewrite(self):
        """
        Rewrite this element and call rewrite on all children.
        Call this base method before rewriting the element.
        """
        del self._error_messages[:]
        for child in self.children():
            child.rewrite()

    @property
    def enabled(self):
        return True

    def get_bypassed(self):
        return False

    ##############################################
    # Tree-like API
    ##############################################
    @property
    def parent(self):
        return self._parent()

    def get_parent_by_type(self, cls):
        parent = self.parent
        if parent is None:
            return None
        elif isinstance(parent, cls):
            return parent
        else:
            return parent.get_parent_by_type(cls)

    @lazy_property
    def parent_platform(self):
        from .platform import Platform
        return self.get_parent_by_type(Platform)

    @lazy_property
    def parent_flowgraph(self):
        from .FlowGraph import FlowGraph
        return self.get_parent_by_type(FlowGraph)

    @lazy_property
    def parent_block(self):
        from .blocks import Block
        return self.get_parent_by_type(Block)

    def reset_parents_by_type(self):
        """Reset all lazy properties"""
        for name, obj in vars(Element):  # explicitly only in Element, not subclasses
            if isinstance(obj, lazy_property):
                delattr(self, name)

    def children(self):
        return
        yield  # empty generator

    ##############################################
    # Type testing
    ##############################################
    is_flow_graph = False
    is_block = False
    is_dummy_block = False
    is_connection = False
    is_port = False
    is_param = False
    is_variable = False
    is_import = False
    is_snippet = False

    def get_raw(self, name):
        descriptor = getattr(self.__class__, name, None)
        if not descriptor:
            raise ValueError("No evaluated property '{}' found".format(name))
        return getattr(self, descriptor.name_raw, None) or getattr(self, descriptor.name, None)

    def set_evaluated(self, name, value):
        descriptor = getattr(self.__class__, name, None)
        if not descriptor:
            raise ValueError("No evaluated property '{}' found".format(name))
        self.__dict__[descriptor.name] = value



from .utils import Spec, expand

MANIFEST_SCHEME = expand(
    title=Spec(types=str, required=True, item_scheme=None),
    version=str,
    brief=str,
    website=str,
    dependencies=list,
    repo=str,
    copyright_owner=list,
    gr_supported_version=list,
    tags=list,
    license=str,
    description=str,
    author=list,
    icon=str,
    file_format=Spec(types=int, required=True, item_scheme=None),
)



"""Copyright 2024 The GNU Radio Contributors
This file is part of GNU Radio

SPDX-License-Identifier: GPL-2.0-or-later
"""


import os
from os.path import expanduser, normpath, expandvars, exists
from collections import OrderedDict

from ..main import get_state_directory, get_config_file_path
from . import Constants


class Config(object):
    name = 'GNU Radio Companion (no gui)'
    license = __doc__.strip()
    website = 'https://www.gnuradio.org/'

    hier_block_lib_dir = os.environ.get('GRC_HIER_PATH', get_state_directory())

    def __init__(self, version, version_parts=None, name=None, prefs=None):
        self._gr_prefs = prefs if prefs else DummyPrefs()
        self.version = version
        self.version_parts = version_parts or version[1:].split(
            '-', 1)[0].split('.')[:3]
        self.enabled_components = self._gr_prefs.get_string(
            'grc', 'enabled_components', '')
        if name:
            self.name = name

    @property
    def block_paths(self):
        paths_sources = (
            self.hier_block_lib_dir,
            os.environ.get('GRC_BLOCKS_PATH', ''),
            self._gr_prefs.get_string('grc', 'local_blocks_path', ''),
            self._gr_prefs.get_string('grc', 'global_blocks_path', ''),
        )

        collected_paths = sum((paths.split(os.pathsep)
                               for paths in paths_sources), [])

        valid_paths = [normpath(expanduser(expandvars(path)))
                       for path in collected_paths if exists(path)]
        # Deduplicate paths to avoid warnings about finding blocks twice, but
        # preserve order of paths
        valid_paths = list(OrderedDict.fromkeys(valid_paths))

        return valid_paths

    @property
    def example_paths(self):
        return [self._gr_prefs.get_string('grc', 'examples_path', '')]

    @property
    def default_flow_graph(self):
        user_default = (
            os.environ.get('GRC_DEFAULT_FLOW_GRAPH') or
            self._gr_prefs.get_string('grc', 'default_flow_graph', '') or
            os.path.join(self.hier_block_lib_dir, 'default_flow_graph.grc')
        )
        return user_default if exists(user_default) else Constants.DEFAULT_FLOW_GRAPH


class DummyPrefs(object):

    def get_string(self, category, item, default):
        return str(default)

    def set_string(self, category, item, value):
        pass

    def get_long(self, category, item, default):
        return int(default)

    def save(self):
        pass



# Copyright 2014-2022 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

# Standard modules
import logging

from qtpy import QtWidgets

# Custom modules
from .. import base

# Logging
log = logging.getLogger(f"grc.application.{__name__}")


class WikiTab(QtWidgets.QDockWidget, base.Component):
    def __init__(self):
        super(WikiTab, self).__init__()

        self.qsettings = self.app.qsettings

        self.setObjectName('wiki_tab')
        self.setWindowTitle('Wiki')

        self.setFloating(False)

        try:
            from qtpy.QtWebEngineWidgets import QWebEngineView
        except ImportError:
            log.error("PyQt QWebEngine missing!")
            self.hide()
            return

        # GUI Widgets

        # Create the layout widget
        container = QtWidgets.QWidget(self)
        container.setObjectName('wiki_tab::container')
        self._container = container

        layout = QtWidgets.QVBoxLayout(container)
        layout.setObjectName('wiki_tab::layout')
        layout.setSpacing(0)
        layout.setContentsMargins(5, 0, 5, 5)
        self._text = QWebEngineView()
        self._text.setZoomFactor(0.5)
        layout.addWidget(self._text)
        self._layout = layout

        container.setLayout(layout)
        self.setWidget(container)

        # TODO: Move to the base controller and set actions as class attributes
        # Automatically create the actions, menus and toolbars.
        # Child controllers need to call the register functions to integrate into the mainwindow
        self.actions = {}
        self.menus = {}
        self.toolbars = {}
        self.createActions(self.actions)
        self.createMenus(self.actions, self.menus)
        self.createToolbars(self.actions, self.toolbars)
        self.connectSlots()

        # Register the dock widget through the AppController.
        # The AppController then tries to find a saved dock location from the preferences
        # before calling the MainWindow Controller to add the widget.
        self.app.registerDockWidget(self, location=self.settings.window.WIKI_TAB_DOCK_LOCATION)
        if not self.qsettings.value("appearance/display_wiki", False, type=bool):
            self.hide()

    def setURL(self, url):
        if not self.isHidden():
            self._text.load(url)
            self._text.show()

    # Actions

    def createActions(self, actions):
        pass

    def createMenus(self, actions, menus):
        pass

    def createToolbars(self, actions, toolbars):
        pass



# Copyright 2014-2020 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

# Standard modules
import abc
import logging
import weakref

# Third-party  modules
from qtpy import QtWidgets


# Logging
log = logging.getLogger(f"grc.application.{__name__}")


class Component(object):
    ''' Abstract base class for all grc components. '''
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        '''
        Initializes the component's base class.
        Sets up the references to the QApplication and the GNU Radio platform.
        Calls createActions() to initialize the component's actions.
        '''
        log.debug("Initializing {}".format(self.__class__.__name__))

        # Application reference - Use weak references to avoid issues with circular references
        # Platform and settings properties are accessed through this reference
        self._app = weakref.ref(QtWidgets.QApplication.instance())

        # Automatically create the actions, menus and toolbars.
        # Child controllers need to call the register functions to integrate into the mainwindow
        self.actions = {}
        self.menus = {}
        self.toolbars = {}
        '''
        self.createActions(self.actions)
        self.createMenus(self.actions, self.menus)
        self.createToolbars(self.actions, self.toolbars)
        '''

        log.debug("Connecting signals")
        self.connectSlots()

    # Properties

    @property
    def app(self):
        return self._app()

    @property
    def settings(self):
        return self._app().settings

    @property
    def platform(self):
        return self._app().platform

    # Required methods

    @abc.abstractmethod
    def createActions(self, actions):
        ''' Add actions to the component. '''
        raise NotImplementedError()

    @abc.abstractmethod
    def createMenus(self, actions, menus):
        ''' Add menus to the component. '''
        raise NotImplementedError()

    @abc.abstractmethod
    def createToolbars(self, actions, toolbars):
        ''' Add toolbars to the component. '''
        raise NotImplementedError()

    # Base methods

    def connectSlots(self, useToggled=True, toggledHandler='_toggled',
                     triggeredHandler="_triggered"):
        '''
         Handles connecting signals from given actions to handlers
            self    - Calling class
            actions - Dictionary of a QAction and unique key

         Dynamically build the connections for the signals by finding the correct
         function to handle an action. Default behavior is to connect checkable actions to
         the 'toggled' signal and normal actions to the 'triggered' signal. If 'toggled' is
         not avaliable or useToggled is set to False then try to connect it to triggered.
         Both toggled and triggered are called for checkable items, so there is no need for
         both to be connected.

            void QAction::toggled ( bool checked ) [signal]
            void QAction::triggered ( bool checked = false ) [signal]
              - Checked is set for checkable actions

         Essentially the same as QMetaObject::connectSlotsByName, except the actions
          and slots can be separated into a view and controller class

        '''
        actions = self.actions
        for key in actions:
            if useToggled and actions[key].isCheckable():
                # Try to use toggled rather than triggered
                try:
                    handler = key + toggledHandler
                    actions[key].toggled.connect(getattr(self, handler))
                    log.debug("<{0}.toggled> connected to handler <{1}>".format(key, handler))
                    # Successful connection. Jump to the next action.
                    continue
                except:
                    # Default to the triggered handler
                    log.warning("Could not connect <{0}.toggled> to handler <{1}>".format(key, handler))

            # Try and bind the 'triggered' signal to a handler.
            try:
                handler = key + triggeredHandler
                actions[key].triggered.connect(getattr(self, handler))
                log.debug("<{0}.triggered> connected to handler <{1}>".format(key, handler))
            except:
                try:
                    log.warning("Handler not implemented for <{0}.triggered> in {1}".format(
                                key, type(self).__name__))
                    actions[key].triggered.connect(getattr(self, 'notImplemented'))
                except:
                    # This should never happen
                    log.error("Class cannot handle <{0}.triggered>".format(key))

    def notImplemented(self):
        log.warning('Not implemented')



import logging

from qtpy.QtGui import QPen, QPainter, QBrush, QFont, QFontMetrics
from qtpy.QtCore import Qt, QPointF, QRectF, QUrl
from qtpy.QtWidgets import QGraphicsItem, QApplication, QAction

from . import colors
from ... import Constants
from ... import Utils
from ....core.blocks.block import Block as CoreBlock
from ....core.utils import flow_graph_complexity

from ..dialogs import PropsDialog

# Logging
log = logging.getLogger(f"grc.application.{__name__}")

LONG_VALUE = 20  # maximum length of a param string.
# if exceeded, '...' will be displayed


class Block(CoreBlock):
    """
    A block. Accesses its graphical representation with self.gui.
    """
    @classmethod
    def make_cls_with_base(cls, super_cls):
        name = super_cls.__name__
        bases = (super_cls,) + cls.__bases__[:-1]
        namespace = cls.__dict__.copy()
        return type(name, bases, namespace)

    def __init__(self, parent, **n):
        super(self.__class__, self).__init__(parent, **n)

        self.width = 50  # will change immediately after the first paint
        self.block_label = self.key

        self.old_data = None

        if "rotation" not in self.states.keys():
            self.states["rotation"] = 0.0

        self.gui = GUIBlock(self, parent)

    def import_data(self, name, states, parameters, **_):
        super(self.__class__, self).import_data(name, states, parameters, **_)
        self.gui.setPos(*self.states["coordinate"])
        self.gui.setRotation(self.states["rotation"])
        self.rewrite()
        self.gui.create_shapes_and_labels()

    def update_bus_logic(self):
        for direc in {'source', 'sink'}:
            if direc == 'source':
                ports = self.sources
                ports_gui = self.filter_bus_port(self.sources)
            else:
                ports = self.sinks
                ports_gui = self.filter_bus_port(self.sinks)
            if 'bus' in map(lambda a: a.dtype, ports):
                for port in ports_gui:
                    self.parent_flowgraph.gui.removeItem(port.gui)
        super(self.__class__, self).update_bus_logic()


class GUIBlock(QGraphicsItem):
    """
    The graphical representation of a block. Accesses the actual block with self.core.
    """

    def __init__(self, core, parent, **n):
        super(GUIBlock, self).__init__()
        self.core = core
        self.parent = self.scene()
        self.font = QFont("Helvetica", 10)

        self.create_shapes_and_labels()

        if "coordinate" not in self.core.states.keys():
            self.core.states["coordinate"] = (500, 300)
            self.setPos(*self.core.states["coordinate"])

        self.old_pos = (self.x(), self.y())
        self.new_pos = (self.x(), self.y())
        self.moving = False

        self.props_dialog = None
        self.right_click_menu = None

        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges)

        self.markups_height = 0.0
        self.markups_width = 0.0
        self.markups = []
        self.markup_text = ""

    def create_shapes_and_labels(self):
        qsettings = QApplication.instance().qsettings
        self.force_show_id = qsettings.value('grc/show_block_ids', type=bool)
        self.hide_variables = qsettings.value('grc/hide_variables', type=bool)
        self.hide_disabled_blocks = qsettings.value('grc/hide_disabled_blocks', type=bool)
        self.snap_to_grid = qsettings.value('grc/snap_to_grid', type=bool)
        self.show_complexity = qsettings.value('grc/show_complexity', type=bool)
        self.show_block_comments = qsettings.value('grc/show_block_comments', type=bool)
        self.show_param_expr = qsettings.value('grc/show_param_expr', type=bool)
        self.show_param_val = qsettings.value('grc/show_param_val', type=bool)
        self.prepareGeometryChange()
        self.font.setBold(True)

        # figure out height of block based on how many params there are
        i = 35

        # Check if we need to increase the height to fit all the ports
        for key, item in self.core.params.items():
            value = item.value
            if (value is not None and item.hide == "none") or (item.dtype == 'id' and self.force_show_id):
                i += 20

        self.height = i

        def get_min_height_for_ports(ports):
            min_height = (
                2 * Constants.PORT_BORDER_SEPARATION +
                len(ports) * Constants.PORT_SEPARATION
            )
            # If any of the ports are bus ports - make the min height larger
            if any([p.dtype == "bus" for p in ports]):
                min_height = (
                    2 * Constants.PORT_BORDER_SEPARATION +
                    sum(
                        port.gui.height + Constants.PORT_SPACING
                        for port in ports
                        if port.dtype == "bus"
                    ) - Constants.PORT_SPACING
                )

            else:
                if ports:
                    min_height -= ports[-1].gui.height
            return min_height

        self.height = max(
            self.height,
            get_min_height_for_ports(self.core.active_sinks),
            get_min_height_for_ports(self.core.active_sources),
        )

        # Figure out width of block based on widest line of text
        fm = QFontMetrics(self.font)
        largest_width = fm.width(self.core.label)
        for key, item in self.core.params.items():
            name = item.name
            value = item.value
            if (value is not None and item.hide == "none") or (item.dtype == 'id' and self.force_show_id):
                if len(value) > LONG_VALUE:
                    value = value[:LONG_VALUE - 3] + '...'

                value_label = item.options[value] if value in item.options else value
                full_line = name + ": " + value_label

                if fm.width(full_line) > largest_width:
                    largest_width = fm.width(full_line)
        self.width = largest_width + 15

        self.markups = []
        self.markup_text = ""
        self.markups_width = 0.0

        if self.show_complexity and self.core.key == "options":
            complexity = flow_graph_complexity.calculate(self.parent)
            self.markups.append('Complexity: {num} bal'.format(
                num=Utils.num_to_str(complexity)))

        if self.show_block_comments and self.core.comment:
            self.markups.append(self.core.comment)

        self.markup_text = "\n".join(self.markups).strip()

        self.markups_height = fm.height() * (self.markup_text.count("\n") + 1)
        for line in self.markup_text.split("\n"):
            if fm.width(line) > self.markups_width:
                self.markups_width = fm.width(line)

        # Update the position and size of all the ports
        bussified = (
            self.core.current_bus_structure["source"],
            self.core.current_bus_structure["sink"],
        )
        for ports, has_busses in zip(
            (self.core.active_sources, self.core.active_sinks), bussified
        ):
            if not ports:
                continue
            port_separation = (
                Constants.PORT_SEPARATION
                if not has_busses
                else ports[0].gui.height + Constants.PORT_SPACING
            )
            offset = (
                self.height - (len(ports) - 1) * port_separation - ports[0].gui.height
            ) / 2
            for port in ports:
                if port._dir == "sink":
                    port.gui.setPos(-15, offset)
                else:
                    port.gui.setPos(self.width, offset)
                port.gui.create_shapes_and_labels()

                offset += (
                    Constants.PORT_SEPARATION
                    if not has_busses
                    else port.gui.height + Constants.PORT_SPACING
                )

        self._update_colors()
        self.create_port_labels()
        self.setTransformOriginPoint(self.width / 2, self.height / 2)

    def create_port_labels(self):
        for ports in (self.core.active_sinks, self.core.active_sources):
            for port in ports:
                port.gui.create_shapes_and_labels()

    def _update_colors(self):
        def get_bg():
            """
            Get the background color for this block
            Explicit is better than a chain of if/else expressions,
            so this was extracted into a nested function.
            """
            if self.core.is_dummy_block:
                return colors.MISSING_BLOCK_BACKGROUND_COLOR
            if self.core.state == "bypassed":
                return colors.BLOCK_BYPASSED_COLOR
            if self.core.state == "enabled":
                if self.core.deprecated:
                    return colors.BLOCK_DEPRECATED_BACKGROUND_COLOR
                return colors.BLOCK_ENABLED_COLOR
            return colors.BLOCK_DISABLED_COLOR

        self._bg_color = get_bg()

    def move(self, x, y):
        self.moveBy(x, y)
        self.core.states["coordinate"] = (self.x(), self.y())

    def paint(self, painter, option, widget):
        if (self.hide_variables and (self.core.is_variable or self.core.is_import)) or (self.hide_disabled_blocks and not self.core.enabled):
            return

        painter.setRenderHint(QPainter.Antialiasing)
        self.font.setBold(True)

        # TODO: Make sure this is correct
        border_color = colors.BORDER_COLOR

        if self.isSelected():
            border_color = colors.HIGHLIGHT_COLOR
        else:
            if self.core.is_dummy_block:
                border_color = colors.MISSING_BLOCK_BORDER_COLOR
            elif self.core.deprecated:
                border_color = colors.BLOCK_DEPRECATED_BORDER_COLOR
            elif self.core.state == "disabled":
                border_color = colors.BORDER_COLOR_DISABLED

        pen = QPen(1)
        pen = QPen(border_color)

        pen.setWidth(3)
        painter.setPen(pen)
        painter.setBrush(QBrush(self._bg_color))
        rect = QRectF(0, 0, self.width, self.height)

        painter.drawRect(rect)
        painter.setPen(QPen(1))

        # Draw block label text
        painter.setFont(self.font)
        if self.core.is_valid():
            painter.setPen(Qt.black)
        else:
            painter.setPen(Qt.red)

        # Adjust the painter if parent block is 180 degrees rotated
        if self.rotation() == 180:
            painter.translate(self.width / 2, self.height / 2)
            painter.rotate(180)
            painter.translate(-self.width / 2, -self.height / 2)

        painter.drawText(
            QRectF(0, 0 - self.height / 2 + 15, self.width, self.height),
            Qt.AlignCenter,
            self.core.label,
        )

        # Draw param text
        y_offset = 30  # params start 30 down from the top of the box
        for key, item in self.core.params.items():
            name = item.name
            value = item.value
            is_evaluated = item.value != str(item.get_evaluated())

            display_value = ""

            # Include the value defined by the user (after evaluation)
            if not is_evaluated or self.show_param_val or not self.show_param_expr:
                display_value += item.options[value] if value in item.options else value  # TODO: pretty_print

            # Include the expression that was evaluated to get the value
            if is_evaluated and self.show_param_expr:
                expr_string = value  # TODO: Truncate

                if display_value:  # We are already displaying the value
                    display_value = expr_string + "=" + display_value
                else:
                    display_value = expr_string

            if len(display_value) > LONG_VALUE:
                display_value = display_value[:LONG_VALUE - 3] + '...'

            value_label = display_value
            if (value is not None and item.hide == "none") or (item.dtype == 'id' and self.force_show_id):
                if item.is_valid():
                    painter.setPen(QPen(1))
                else:
                    painter.setPen(Qt.red)

                self.font.setBold(True)
                painter.setFont(self.font)
                painter.drawText(
                    QRectF(7.5, 0 + y_offset, self.width, self.height),
                    Qt.AlignLeft,
                    name + ": ",
                )
                fm = QFontMetrics(self.font)
                self.font.setBold(False)
                painter.setFont(self.font)
                painter.drawText(
                    QRectF(
                        7.5 + fm.width(name + ": "),
                        0 + y_offset,
                        self.width,
                        self.height,
                    ),
                    Qt.AlignLeft,
                    value_label,
                )
                y_offset += 20

        if self.markup_text:
            painter.setPen(Qt.gray)
            painter.drawText(
                QRectF(0, self.height + 5, self.markups_width, self.markups_height),
                Qt.AlignLeft,
                self.markup_text,
            )

    def boundingRect(self):
        # TODO: Comments should be a separate QGraphicsItem
        return QRectF(
            -2.5, -2.5, self.width + 5 + self.markups_width, self.height + 5 + self.markups_height
        )

    def set_states(self, states):
        for k, v in states.items():
            self.core.states[k] = v

        self.setPos(*self.core.states["coordinate"])
        self.setRotation(self.core.states["rotation"])

    def mousePressEvent(self, e):
        super(self.__class__, self).mousePressEvent(e)
        log.debug(f"{self} clicked")
        url_prefix = str(self.core.parent_platform.config.wiki_block_docs_url_prefix)
        QApplication.instance().WikiTab.setURL(QUrl(url_prefix + self.core.label.replace(" ", "_")))

        self.moveToTop()

    def contextMenuEvent(self, e):
        if not self.isSelected():
            self.scene().clearSelection()
            self.setSelected(True)

        self.right_click_menu = self.scene()._app().MainWindow.menus["edit"]
        example_action = QAction("Examples...")
        example_action.triggered.connect(self.view_examples)
        self.right_click_menu.addAction(example_action)
        self.right_click_menu.exec_(e.screenPos())

    def view_examples(self):
        self.scene().app.MainWindow.example_browser_triggered(key_filter=self.core.key)

    def mouseDoubleClickEvent(self, e):
        self.open_properties()
        super(self.__class__, self).mouseDoubleClickEvent(e)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange and self.scene() and self.snap_to_grid:
            grid_size = 10
            value.setX(round(value.x() / grid_size) * grid_size)
            value.setY(round(value.y() / grid_size) * grid_size)
            return value
        else:
            return QGraphicsItem.itemChange(self, change, value)

    def rotate(self, rotation):
        log.debug(f"Rotating {self.core.name}")
        new_rotation = (self.rotation() + rotation) % 360
        self.setRotation(new_rotation)
        self.core.states["rotation"] = new_rotation
        self.create_shapes_and_labels()
        for con in self.core.connections():
            con.gui.set_rotation(new_rotation)
            con.gui.update()

    def moveToTop(self):
        # TODO: Is there a simpler way to do this?
        self.setZValue(self.scene().getMaxZValue() + 1)

    def center(self):
        return QPointF(self.x() + self.width / 2, self.y() + self.height / 2)

    def open_properties(self):
        self.props_dialog = PropsDialog(self.core, self.force_show_id)
        self.props_dialog.show()



# Copyright 2008-2017 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# SPDX-License-Identifier: GPL-2.0-or-later
#


class TemplateArg(str):
    """
    A cheetah template argument created from a param.
    The str of this class evaluates to the param's to code method.
    The use of this class as a dictionary (enum only) will reveal the enum opts.
    The __call__ or () method can return the param evaluated to a raw python data type.
    """

    def __new__(cls, param):
        value = param.to_code()
        instance = str.__new__(cls, value)
        setattr(instance, '_param', param)
        return instance

    def __getitem__(self, item):
        return str(self._param.get_opt(item)) if self._param.is_enum() else NotImplemented

    def __getattr__(self, item):
        if not self._param.is_enum():
            raise AttributeError()
        try:
            return str(self._param.get_opt(item))
        except KeyError:
            raise AttributeError()

    def __str__(self):
        return str(self._param.to_code())

    def __call__(self):
        return self._param.get_evaluated()



import logging


class GRCHandler(logging.Handler):  # Inherit from logging.Handler
    ''' Custom log handler for GRC. Stores log entries to be viewed using the GRC debug window. '''

    def __init__(self, maxLength=256):
        # run the regular Handler __init__
        logging.Handler.__init__(self)
        # Our custom argument
        self.log = collections.deque(maxlen=maxLength)

    def emit(self, record):
        self.log.append(record)

    def getLogs(self, level):
        pass


class ConsoleFormatter(logging.Formatter):
    '''
     Custom log formatter that nicely truncates the log message and log levels
      - Verbose mode outputs: time, level, message, name, filename, and line number
      - Normal mode output varies based on terminal size:
            w < 80       - Level, Message (min length 40)
            80 < w < 120 - Level, Message, File, Line (25)
            120 < w      - Level, Message, Name, File, Line
      - Color mode ouptuts the same variable sizes and uses the blessings module
        to add color
    '''

    # TODO: Better handle multi line messages. Need to indent them or something

    def __init__(self, verbose=False, short_level=True):
        # Test for blessings formatter
        try:
            from blessings import Terminal
            self.terminal = Terminal()
            self.formatLevel = self.formatLevelColor
        except:
            self.terminal = None
            self.formatLevel = self.formatLevelPlain

        # Default to short
        self.formatLevelLength = self.formatLevelShort
        if not short_level:
            self.formatLevelLength = self.formatLevelLong

        # Setup the format function as a pointer to the correct formatting function
        # Determine size and mode
        # TODO: Need to update the sizes depending on short or long outputs
        import shutil
        size = shutil.get_terminal_size()
        width = max(40, size.columns - 10)
        if size.columns < 80:
            self.format = self.short
            self.width = width
        elif size.columns < 120:
            self.format = self.medium
            self.width = width - 30
        elif size.columns >= 120:
            self.format = self.long
            self.width = width - 45
        # Check if verbose mode. If so override other options
        if verbose:
            self.format = self.verbose

    # Normal log formmatters
    def short(self, record):
        message = self.formatMessage(record.msg, self.width)
        level = self.formatLevel(record.levelname)
        return "{0} -- {1}".format(level, message)

    def medium(self, record):
        message = self.formatMessage(record.msg, self.width)
        level = self.formatLevel(record.levelname)
        output = '{0} -- {1:<' + str(self.width) + '} ({2}:{3})'
        return output.format(level, message, record.filename, record.lineno)

    def long(self, record):
        message = self.formatMessage(record.msg, self.width)
        level = self.formatLevel(record.levelname)
        output = '{0} -- {1:<' + str(self.width) + '} {2} ({3}:{4})'
        return output.format(level, message, record.name, record.filename, record.lineno)

    ''' Verbose formatter '''

    def verbose(self, record):
        # TODO: Still need to implement this
        pass

    ''' Level and message formatters '''
    # Nicely format the levelname
    # Level name can be formated to either short or long, and also with color

    def formatLevelColor(self, levelname):
        term = self.terminal
        output = "{0}{1}{2}{3}"
        level = self.formatLevelLength(levelname)
        if levelname == "DEBUG":
            return output.format(term.blue, "", level, term.normal)
        elif levelname == "INFO":
            return output.format(term.green, "", level, term.normal)
        elif levelname == "WARNING":
            return output.format(term.yellow, "", level, term.normal)
        elif levelname == "ERROR":
            return output.format(term.red, term.bold, level, term.normal)
        elif levelname == "CRITICAL":
            return output.format(term.red, term.bold, level, term.normal)
        else:
            return output.format(term.blue, "", level, term.normal)

    def formatLevelPlain(self, levelname):
        ''' Format the level name without color. formatLevelLength points to the right function '''
        return self.formatLevelLength(levelname)

    def formatLevelShort(self, levelname):
        return "[{0}]".format(levelname[0:1])

    def formatLevelLong(self, levelname):
        output = "{0:<10}"
        if levelname in ["DEBUG", "INFO", "WARNING"]:
            return output.format("[{0}]".format(levelname.capitalize()))
        else:
            return output.format("[{0}]".format(levelname.upper()))

    def formatMessage(self, message, width):
        # First, strip out any newline for console output
        message = message.rstrip()
        if len(message) > width:
            return (message[:(width - 3)] + "...")
        return message



# Copyright 2014-2020 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


# Standard modules
import logging

from qtpy.QtCore import QUrl, Qt, QVariant
from qtpy.QtWidgets import (QLineEdit, QTreeView, QMenu, QDockWidget, QWidget,
                            QAction, QVBoxLayout, QAbstractItemView, QCompleter, QToolButton)
from qtpy.QtGui import QStandardItem, QStandardItemModel

# Custom modules
from .. import base
from .canvas.block import Block

# Logging
log = logging.getLogger(f"grc.application.{__name__}")


class BlockSearchBar(QLineEdit):
    def __init__(self, parent):
        QLineEdit.__init__(self)
        QLineEdit.setClearButtonEnabled(self, True)
        self.parent = parent
        self.setObjectName("block_library::search_bar")
        # We must use a queued connection here, otherwise
        # the searchbar entry will not be deleted if we selected an item from
        # the popup window by key_down and key_enter
        self.returnPressed.connect(self.on_return_pressed, Qt.QueuedConnection)
        # Find clear Button inside QLineEdit
        clr_btn = self.findChildren(QToolButton)
        if len(clr_btn) > 0:
            clr_btn[0].triggered.connect(self.clr)

    def clr(self):
        self.parent.reset()

    def first_block_in_results(self):
        model = self.completer().completionModel()
        index = model.index(0, 0)
        return model.data(index)

    def on_return_pressed(self):
        block_key = None
        label = self.text()
        if label in self.parent._block_tree_flat:
            block_key = self.parent._block_tree_flat[label].key
        else:
            label = self.first_block_in_results()
            if label in self.parent._block_tree_flat:
                block_key = self.parent._block_tree_flat[label].key
        if block_key:
            self.parent.add_block(block_key)
        else:
            log.info(f"No block named {label}")
        self.clr()


def get_items(model):
    items = []
    for i in range(0, model.rowCount()):
        index = model.index(i, 0)
        items.append(model.data(index))
    return items


class LibraryView(QTreeView):
    def __init__(self, parent):
        QTreeView.__init__(self, parent)
        self.library = parent.parent()
        self.contextMenu = QMenu()
        self.example_action = QAction("Examples...")
        self.add_to_fg_action = QAction("Add to flowgraph")
        self.example_action.triggered.connect(self.view_examples)
        self.add_to_fg_action.triggered.connect(self.add_block)
        self.contextMenu.addAction(self.example_action)
        self.contextMenu.addAction(self.add_to_fg_action)

    # TODO: Use selectionChanged() or something instead
    # so we can use arrow keys too
    def updateDocTab(self):
        label = self.model().data(self.currentIndex())
        if label in self.parent().parent()._block_tree_flat:
            prefix = str(
                self.parent().parent().platform.config.wiki_block_docs_url_prefix
            )
            self.parent().parent().app.WikiTab.setURL(
                QUrl(prefix + label.replace(" ", "_"))
            )

    def contextMenuEvent(self, event):
        key = self.model().data(self.currentIndex(), Qt.UserRole)
        if key:  # Modules and categories don't have UserRole data
            self.contextMenu.exec_(self.mapToGlobal(event.pos()))

    def view_examples(self):
        key = self.model().data(self.currentIndex(), Qt.UserRole)
        self.library.app.MainWindow.example_browser_triggered(key_filter=key)

    def add_block(self) -> None:
        key = self.model().data(self.currentIndex(), Qt.UserRole)
        self.library.add_block(key)


class BlockLibrary(QDockWidget, base.Component):
    def __init__(self):
        super(BlockLibrary, self).__init__()

        self.qsettings = self.app.qsettings
        self.blocklibrary_expanded = self.qsettings.value('grc/blocklibrary_expanded', type=bool)

        self.setObjectName("block_library")
        self.setWindowTitle("Block Library")

        # TODO: Pull from preferences and revert to default if not found?
        self.resize(400, 300)
        self.setFloating(False)

        # GUI Widgets

        # Create the layout widget
        container = QWidget(self)
        container.setObjectName("block_library::container")
        self._container = container

        layout = QVBoxLayout(container)
        layout.setObjectName("block_library::layout")
        layout.setSpacing(0)
        layout.setContentsMargins(5, 0, 5, 5)
        self._layout = layout

        # Setup the model for holding block data
        self._model = QStandardItemModel()

        library = LibraryView(container)
        library.setObjectName("block_library::library")
        library.setModel(self._model)
        library.setDragEnabled(True)
        library.setDragDropMode(QAbstractItemView.DragOnly)
        # library.setColumnCount(1)
        library.setHeaderHidden(True)
        # Expand categories with a single click is build in
        library.selectionModel().selectionChanged.connect(library.updateDocTab)
        # library.headerItem().setText(0, "Blocks")
        library.doubleClicked.connect(
            lambda block: self.add_block(block.data(Qt.UserRole))
        )
        self._library = library

        search_bar = BlockSearchBar(self)
        search_bar.setPlaceholderText("Find a block")
        self._search_bar = search_bar

        # Add widgets to the component
        layout.addWidget(search_bar)
        layout.addSpacing(5)
        layout.addWidget(library)
        container.setLayout(layout)
        self.setWidget(container)

        # Translation support

        # self.setWindowTitle(_translate("blockLibraryDock", "Library", None))
        # library.headerItem().setText(0, _translate("blockLibraryDock", "Blocks", None))
        # QMetaObject.connectSlotsByName(blockLibraryDock)

        # Loading blocks

        # Keep as a separate function so it can be called at a later point (Reloading blocks)
        self._block_tree_flat = {}
        self.load_blocks()
        self.populate_tree(self._block_tree)

        completer = QCompleter(self._block_tree_flat.keys())
        completer.setCompletionMode(QCompleter.PopupCompletion)
        completer.setModelSorting(QCompleter.CaseInsensitivelySortedModel)
        completer.setCaseSensitivity(Qt.CaseInsensitive)
        completer.setFilterMode(Qt.MatchContains)
        self._search_bar.setCompleter(completer)
        self._search_bar.textChanged.connect(self.show_search_results)

        # TODO: Move to the base controller and set actions as class attributes
        # Automatically create the actions, menus and toolbars.
        # Child controllers need to call the register functions to integrate into the mainwindow
        self.actions = {}
        self.menus = {}
        self.toolbars = {}
        self.createActions(self.actions)
        self.createMenus(self.actions, self.menus)
        self.createToolbars(self.actions, self.toolbars)
        self.connectSlots()

        # Register the dock widget through the AppController.
        # The AppController then tries to find a saved dock location from the preferences
        # before calling the MainWindow Controller to add the widget.
        self.app.registerDockWidget(
            self, location=self.settings.window.BLOCK_LIBRARY_DOCK_LOCATION
        )

        # Register the menus
        # self.app.registerMenu(self.menus["library"])

        # Dict representing which examples contain various blocks
        self.examples_w_block = {}

        if not self.qsettings.value("appearance/display_blocklibrary", True, type=bool):
            self.hide()

    def createActions(self, actions):
        pass

    def createMenus(self, actions, menus):
        pass

    def createToolbars(self, actions, toolbars):
        pass

    def show_search_results(self):
        if self._search_bar.text() == '':
            self.populate_tree(self._block_tree)
        else:
            self.populate_tree(self._block_tree, get_items(self._search_bar.completer().completionModel()))

    def reset(self):
        """Reset the filter (show all blocks in the tree view)"""
        self._search_bar.clear()
        self.populate_tree(self._block_tree)

    def load_blocks(self):
        """Load the block tree from the platform and populate the widget."""
        # Loop through all of the blocks and create the nested hierarchy (this can be unlimited nesting)
        # This takes advantage of Python's use of references to move through the nested layers

        log.info("Loading blocks")
        block_tree = {}
        for block in self.platform.blocks.values():
            if block.category:
                # Blocks with None category should be left out for whatever reason (e.g. not installed)
                # print(block.category) # in list form, e.g. ['Core', 'Digital Television', 'ATSC']
                # print(block.label) # label GRC uses to name block
                # print(block.key) # actual block name (i.e. class name)

                # Create a copy of the category list so things can be removed without changing the original list
                category = block.category[:]

                # Get a reference to the main block tree.
                # As nested categories are added, this is updated to point to the proper sub-tree in the next layer
                sub_tree = block_tree
                while category:
                    current = category.pop(0)
                    if current not in sub_tree.keys():
                        # Create the new sub-tree
                        sub_tree[current] = {}
                    # Move to the next layer in the block tree
                    sub_tree = sub_tree[current]
                # Sub_tree should now point at the final node of the block_tree that contains the block
                # Add a reference to the block object to the proper subtree
                sub_tree[block.label] = block
                self._block_tree_flat[block.label] = block
        # Save a reference to the block tree in case it is needed later
        self._block_tree = block_tree

    def reload_blocks(self):
        self.platform.build_library()
        self.load_blocks()
        self.populate_tree(self._block_tree)

    def add_block(self, block_key: str):
        """Add a block by its key."""
        if block_key is None:
            return

        scene = self.app.MainWindow.currentFlowgraphScene
        view = self.app.MainWindow.currentView
        pos_ = view.mapToScene(view.viewport().rect().center())
        scene.add_block(block_key, pos=(pos_.x(), pos_.y()))

    def populate_tree(self, block_tree, v_blocks=None):
        """Populate the item model and tree view with the hierarchical block tree."""
        # Recursive method of populating the QStandardItemModel
        # Since the _model.invisibleRootItem is the initial parent, this will populate
        # the model which is used for the TreeView.
        self._model.removeRows(0, self._model.rowCount())

        def _populate(blocks, parent):
            found = False
            for name, obj in sorted(blocks.items()):
                child_item = QStandardItem()
                child_item.setEditable(False)
                if type(obj) is dict:  # It's a category
                    child_item.setText(name)
                    child_item.setDragEnabled(
                        False
                    )  # categories should not be draggable
                    if not _populate(obj, child_item):
                        continue
                    else:
                        found = True
                else:  # It's a block
                    if v_blocks and not name in v_blocks:
                        continue
                    else:
                        found = True
                    child_item.setText(obj.label)
                    child_item.setDragEnabled(True)
                    child_item.setSelectable(True)
                    child_item.setData(
                        QVariant(obj.key),
                        role=Qt.UserRole,
                    )
                parent.appendRow(child_item)
            return found

        # Call the nested function recursively to populate the block tree
        log.debug("Populating the treeview")
        _populate(block_tree, self._model.invisibleRootItem())
        if self.blocklibrary_expanded:
            for row in range(0, self._model.rowCount()):
                self._library.expand(self._model.item(row, 0).index())
        else:
            self._library.expand(
                self._model.item(0, 0).index()
            )
        if v_blocks:
            self._library.expandAll()

    def populate_w_examples(self, examples_w_block: dict[str, set[str]], designated_examples_w_block: dict[str, set[str]]):
        """
        Store the examples in the block library.
        See the ExampleBrowser for more info.
        """
        self.examples_w_block = examples_w_block
        self.designated_examples_w_block = designated_examples_w_block

    def get_examples(self, block_key: str) -> list[Block]:
        """Get the example flowgraphs that contain a certain block"""
        try:
            return self.designated_examples_w_block[block_key]
        except:
            if block_key in self.examples_w_block:
                return self.examples_w_block[block_key]
            else:
                return []
